/*
 * This file is a part of AllPlaceFinder library
 * and is distributed under the 2-Clause BSD licence.
 * Author: Tomasz Sowa <t.sowa@ttmath.org>
 */

/*
 * Copyright (c) 2021, Tomasz Sowa
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <iostream>
#include "mainoptions/mainoptionsparser.h"
#include "finder.h"
#include "params.h"


void print_syntax()
{
	std::cout << "placefinder options:" << std::endl;
	std::cout << " -h print this help" << std::endl;
	std::cout << " --osm-file file_name - a path to osm file name" << std::endl;
	std::cout << " --lookup-dir dir_name - a path to lookup dir" << std::endl;
	std::cout << " --create-lookup-dir - create a lookup dir, you have to provide a path in --lookup-dir parameter" << std::endl;
}



bool parse_parametr(const pt::Space * input_par, std::wstring & param, const char * error_msg)
{
	if( input_par )
	{
		if( input_par->is_table() && input_par->table_size() == 1  )
		{
			pt::Space * first_item = input_par->value.value_table.back();

			if( first_item->is_table() && first_item->table_size() == 1 )
				param = first_item->value.value_table.back()->to_wstr();
		}
		else
		{
			std::cout << error_msg << std::endl;
			return false;
		}
	}

	return true;
}


bool parse_parameters(const pt::Space & parameters, allplacefinder::Params & params)
{
	const pt::Space * osm_file = parameters.get_space(L"osm-file");
	const pt::Space * lookup_dir = parameters.get_space(L"lookup-dir");
	const pt::Space * create_lookup_dir = parameters.get_space(L"create-lookup-dir");
	bool ok = true;

	if( osm_file )
	{
		ok = ok && parse_parametr(osm_file, params.osm_file, "--osm-file parametr can be used only once");
	}

	if( lookup_dir )
	{
		ok = ok && parse_parametr(lookup_dir, params.lookup_dir, "--lookup-dir parametr can be used only once");
	}

	params.create_lookup_dir = (create_lookup_dir != nullptr);
	return ok;
}


int make_action(allplacefinder::Params & params)
{
	int status = 3;

	if( params.create_lookup_dir )
	{
		allplacefinder::Finder finder;
		status = finder.create_lookup_dir(params) ? 0 : 4;
	}

	return status;
}



int main(int argc, const char ** argv)
{
pt::Space options, arguments;
pt::MainOptionsParser options_parser;
allplacefinder::Params params;

	arguments.add(L"osm-file", 1);
	arguments.add(L"lookup-dir", 1);

	pt::MainOptionsParser::Status status = options_parser.parse(argc, argv, options, arguments);

	if( status != pt::MainOptionsParser::status_ok )
	{
		print_syntax();
		return 1;
	}

	if( options.has_key(L"h") )
	{
		print_syntax();
		return 0;
	}

	if( parse_parameters(options, params) )
	{
		// ------------------
		// for debug purposes
		//params.create_lookup_dir = true;

		//params.osm_file = L"/home/tomek/data/test.xml";
		//params.osm_file = L"/home/tomek/data/wroclaw_map.xml";

		//params.lookup_dir = L"/home/tomek/data/lookup-dir";
		//params.lookup_dir = L"/data/lookup-dir";
		// ------------------

		return make_action(params);
	}

	//std::cout << options.serialize_to_json_str(true) << std::endl;

return 2;
}

